import json
from typing import Tuple
from uuid import uuid4

from mypy_boto3_s3.service_resource import Bucket
from mypy_boto3_sqs.service_resource import Queue

from nhs_aws_helpers import events_client

from .test_helpers import read_sqs_json_messages


def test_get_temp_queue(temp_queue: Queue):
    assert temp_queue
    assert temp_queue.url

    messages = read_sqs_json_messages(temp_queue)
    assert not messages

    temp_queue.send_message(MessageBody=json.dumps({"my_event": 1234}))

    messages = read_sqs_json_messages(temp_queue)
    assert messages
    assert len(messages) == 1
    assert messages[0]["my_event"] == 1234


def test_get_temp_fifo_queue(temp_fifo_queue: Queue):
    assert temp_fifo_queue
    assert temp_fifo_queue.url

    messages = read_sqs_json_messages(temp_fifo_queue)
    assert not messages

    temp_fifo_queue.send_message(
        MessageGroupId="testing", MessageDeduplicationId=uuid4().hex, MessageBody=json.dumps({"my_event": 1234})
    )

    messages = read_sqs_json_messages(temp_fifo_queue)
    assert messages
    assert len(messages) == 1
    assert messages[0]["my_event"] == 1234


def test_get_temp_event_bus(temp_event_bus: Tuple[Queue, str]):
    events_queue, bus_name = temp_event_bus
    assert events_queue
    assert events_queue.url
    event_bus = events_client()

    messages = read_sqs_json_messages(events_queue)
    assert not messages

    event_bus.put_events(
        Entries=[
            {
                "Source": "my_source",
                "DetailType": "SuperEvent",
                "EventBusName": bus_name,
                "Detail": json.dumps({"my_event": 1234}),
            }
        ]
    )

    messages = read_sqs_json_messages(events_queue)
    assert messages
    assert len(messages) == 1
    assert messages[0]["detail"]["my_event"] == 1234


def test_get_temp_s3_bucket(temp_s3_bucket: Bucket):
    assert temp_s3_bucket
    assert temp_s3_bucket.name

    existing_objects = list(temp_s3_bucket.objects.all())
    assert not existing_objects
    key = uuid4().hex
    value = uuid4().hex.encode()
    new_object = temp_s3_bucket.Object(key)
    new_object.put(Body=value)

    all_objects = list(temp_s3_bucket.objects.all())
    assert len(all_objects) == 1
    got_object = temp_s3_bucket.Object(key)
    res = got_object.get()
    assert res["Body"].read() == value


def test_get_temp_s3_bucket_again_to_check_empty(temp_s3_bucket: Bucket):
    assert temp_s3_bucket
    assert temp_s3_bucket.name

    existing_objects = list(temp_s3_bucket.objects.all())
    assert not existing_objects



import sys
from io import SEEK_CUR, SEEK_END, BytesIO
from time import time
from typing import Any, List, cast
from uuid import uuid4

import pytest
from mypy_boto3_s3.service_resource import Bucket, Object

from nhs_aws_helpers import s3_object
from nhs_aws_helpers.async_s3_object_reader import AsyncS3ObjectReader
from nhs_aws_helpers.async_s3_object_writer import AsyncS3ObjectWriter
from nhs_aws_helpers.common import MiB

_DEFAULT_BUFFER_SIZE = 200

_TEXT = [
    "It bégan with the forging of the Great Rings.\r\n",
    "Threé were given to the Elves, immortal, wisest and fairest of all beings.\n",
    "Seven to the Dwarf-Lords, great miners and craftsmen of the mountain halls.\r\n",
    "And nine, nine rings were gifted to the race of Men, who above all else desire power.\r\n",
    "For within these rings was bound the strength and the will to govern each race.\r\n",
    "But théy were all of them deceived, for another ring was made.\r\n",
    "Déép in the land of Mordor, in thé Fires of Mount Doom, thé Dark Lord Sauron forged a master "
    "ring in secret, and into this ring he poured his cruelty, his malice and his will to dominaté "
    "all life.\n",
    "Oné ring to rule them all.",
]
_S3_TEXT_BODY = "".join(_TEXT)
_TEXT_BODY_LEN = len(_S3_TEXT_BODY)

_ASCII_TEXT = [
    "It began with the forging of the Great Rings.\r\n",
    "Three were given to the Elves, immortal, wisest and fairest of all beings.\n",
    "Seven to the Dwarf-Lords, great miners and craftsmen of the mountain halls.\r\n",
    "And nine, nine rings were gifted to the race of Men, who above all else desire power.\r\n",
    "For within these rings was bound the strength and the will to govern each race.\r\n",
    "But they were all of them deceived, for another ring was made.\r\n",
    "Deep in the land of Mordor, in the Fires of Mount Doom, the Dark Lord Sauron forged a master "
    "ring in secret, and into this ring he poured his cruelty, his malice and his will to dominate "
    "all life.\n",
    "One ring to rule them all.",
]
_ASCII_S3_TEXT_BODY = "".join(_ASCII_TEXT)
_ASCII_TEXT_BODY_LEN = len(_ASCII_S3_TEXT_BODY)

_NON_ASCII_TEXT = [
    "It begдn with the forging of the Great Rings.\r\n",
    "Three were given to the Elves, immortдl, wisest and fairest of дll beings.\n",
    "Seven to the Dwarf-Lords, great miners and craftsmen of the mountaiи halls.\r\n",
    "And nine, nine rings were gifted to the race of Men, who above all else desiлe power.\r\n",
    "For witлin these rings was bound the strength and the will to govern each race.\r\n",
    "But they were all of them deceived, for another ring was made.\r\n",
    "Deep in the land of Mordor, in the Fires of Mount Doom, the Dark Lord Sauлon forged a master "
    "ring in secret, and into this ring he poured his cruelty, his malice and his will to dominate "
    "all life.\n",
    "One ring to rule them all. Владимир Путин",
]
_NON_ASCII_S3_TEXT_BODY = "".join(_NON_ASCII_TEXT)
_NON_ASCII_TEXT_BODY_LEN = len(_NON_ASCII_S3_TEXT_BODY)
_NON_ASCII_TEXT_BODY_LEN_ENCODED = len(_NON_ASCII_S3_TEXT_BODY.encode("utf-8"))

_BINARY = b"Example bytes"


class MockS3Object:
    def __init__(self, bytes_io):
        self.bytes_io = bytes_io

    def get(self):
        return {"Body": self.bytes_io}

    def shh(self):
        return self.__dict__


def s3_reader_text():
    bytes_io = BytesIO(_S3_TEXT_BODY.encode("utf-8"))

    s3_io = AsyncS3ObjectReader(
        s3_obj=cast(Object, MockS3Object(bytes_io)), buffer_size=_DEFAULT_BUFFER_SIZE, encoding="utf-8"
    )

    return s3_io


def ascii_s3_reader_text():
    bytes_io = BytesIO(_ASCII_S3_TEXT_BODY.encode("utf-8"))

    s3_io = AsyncS3ObjectReader(
        s3_obj=cast(Object, MockS3Object(bytes_io)), buffer_size=_DEFAULT_BUFFER_SIZE, encoding="utf-8"
    )

    return s3_io


def non_ascii_s3_reader_text():
    bytes_io = BytesIO(_NON_ASCII_S3_TEXT_BODY.encode("utf-8"))

    s3_io = AsyncS3ObjectReader(
        s3_obj=cast(Object, MockS3Object(bytes_io)), buffer_size=_DEFAULT_BUFFER_SIZE, encoding="utf-8"
    )

    return s3_io


def s3_reader_binary():
    bytes_io = BytesIO(_BINARY)

    s3_io = AsyncS3ObjectReader(s3_obj=cast(Object, MockS3Object(bytes_io)), buffer_size=_DEFAULT_BUFFER_SIZE)

    return s3_io


def s3_writer(bucket: Bucket, key: str):
    return AsyncS3ObjectWriter(s3_object(bucket.name, key), encoding="utf-8")


def s3_writer_binary(bucket: Bucket, key: str):
    return AsyncS3ObjectWriter(s3_object(bucket.name, key))


async def test_read_all() -> None:
    async with s3_reader_text() as reader:
        body = await reader.read()

    assert body == _S3_TEXT_BODY


async def test_read_n() -> None:
    size = 25

    chunks = []

    async with s3_reader_text() as reader:
        chunk = await reader.read(size)

        pos = 0

        assert len(reader._buffer.encode("utf-8")) == _DEFAULT_BUFFER_SIZE

        while chunk:
            chunks.append(chunk)
            actual = _S3_TEXT_BODY[pos : pos + len(chunk)]
            assert chunk == actual
            pos += len(chunk)
            chunk = await reader.read(size)

    assert "".join(chunks) == _S3_TEXT_BODY


async def test_read_negative() -> None:
    async with s3_reader_text() as reader:
        first_chunk = await reader.read(300)

        neg_chunk = await reader.read(-33)

        assert _S3_TEXT_BODY.encode("utf-8")[0:300].decode("utf-8") == first_chunk

        assert first_chunk[-33:] == neg_chunk

        assert reader._bytes_read == 298

        assert reader._position == 265


async def test_read_negative_all() -> None:
    read_size = -57

    async with s3_reader_text() as reader:
        _ = await reader.read()
        chunks: List[str] = []
        chunk = await reader.read(read_size)
        while chunk:
            chunks.insert(0, chunk)
            chunk = await reader.read(read_size)
        joined = "".join(chunks)
        assert joined == _S3_TEXT_BODY


async def test_seek_back_and_forth() -> None:
    async with non_ascii_s3_reader_text() as reader:
        pos = await reader.seek(57)

        assert pos == 57
        assert reader._position == 57
        assert reader._bytes_read <= _DEFAULT_BUFFER_SIZE

        pos = await reader.seek(57)

        assert pos == 57
        assert reader._position == 57
        assert reader._bytes_read <= _DEFAULT_BUFFER_SIZE

        new_pos = _DEFAULT_BUFFER_SIZE + 57
        pos = await reader.seek(new_pos)
        assert new_pos == pos
        assert new_pos == reader._position
        assert reader._bytes_read <= 2 * _DEFAULT_BUFFER_SIZE

        pos = await reader.seek(_NON_ASCII_TEXT_BODY_LEN_ENCODED + 1)

        assert pos == _NON_ASCII_TEXT_BODY_LEN

        pos = await reader.seek(100, SEEK_END)

        assert pos == _NON_ASCII_TEXT_BODY_LEN - 100


async def test_seek_relative() -> None:
    pos = 0

    async with non_ascii_s3_reader_text() as reader:
        for i in range(7):
            pos = await reader.seek(100, SEEK_CUR)

            assert pos == min((i + 1) * 100, _NON_ASCII_TEXT_BODY_LEN)

        assert pos == _NON_ASCII_TEXT_BODY_LEN

        for i in range(7):
            pos = await reader.seek(-100, SEEK_CUR)

            assert pos == max(0, _NON_ASCII_TEXT_BODY_LEN - ((i + 1) * 100))

        assert pos == 0

        read_all = await reader.read()

        assert read_all == _NON_ASCII_S3_TEXT_BODY


async def test_read_line() -> None:
    async with s3_reader_text() as reader:
        line = await reader.readline()

        assert line == _TEXT[0]

        line = await reader.readline()

        assert line == _TEXT[1]

        for _ in range(len(_TEXT) - 2):
            line = await reader.readline()

        assert line == _TEXT[-1]


async def test_read_line_on_binary_object():
    with pytest.raises(ValueError, match="readline only works if _encoding was defined"):
        async with s3_reader_binary() as reader:
            await reader.readline()


async def test_read_lines() -> None:
    async with s3_reader_text() as reader:
        line = await reader.readline()

        assert (_TEXT[0]) == line

        lines = [line async for line in reader.readlines()]

        assert (_TEXT[1:]) == lines


if sys.version_info < (3, 10):

    async def anext(gen, default: Any = None):
        async for el in gen:
            return el
        return default


async def test_read_lines_on_binary_object():
    with pytest.raises(ValueError, match="readlines only works if _encoding was defined"):
        async with s3_reader_binary() as reader:
            await anext(reader.readlines())  # type: ignore[name-defined]


async def aenumerate(asequence, start=0):
    """Asynchronously enumerate an async iterator from a given start value"""
    num = start
    async for elem in asequence:
        yield num, elem
        num += 1


async def test_iter() -> None:
    async with s3_reader_text() as reader:
        async for idx, a_line in aenumerate(reader.readlines()):
            assert a_line.strip() == _TEXT[idx].strip()


async def test_write_binary(temp_s3_bucket) -> None:
    key = f"testkey/{uuid4().hex}"
    async with s3_writer_binary(temp_s3_bucket, key) as writer:
        await writer.write(_BINARY)
        await writer.close()

    read = temp_s3_bucket.Object(key).get()["Body"].read()

    assert read == _BINARY


async def test_write_binary_then_nothing(temp_s3_bucket) -> None:
    key = f"testkey/{uuid4().hex}"
    async with s3_writer_binary(temp_s3_bucket, key) as writer:
        await writer.write(_BINARY)
        await writer.write(b"")
        await writer.close()

    read = temp_s3_bucket.Object(key).get()["Body"].read()

    assert read == _BINARY


async def test_write_string(temp_s3_bucket) -> None:
    key = f"testkey/{uuid4().hex}"
    async with s3_writer(temp_s3_bucket, key) as writer:
        await writer.writelines(_S3_TEXT_BODY)
        await writer.close()

    read = temp_s3_bucket.Object(key).get()["Body"].read()
    decoded = read.decode("utf-8")
    assert decoded == _S3_TEXT_BODY


async def test_write_lines_string(temp_s3_bucket) -> None:
    key = f"testkey/{uuid4().hex}"
    async with s3_writer(temp_s3_bucket, key) as writer:
        await writer.writelines(_TEXT)
        await writer.close()

    read = temp_s3_bucket.Object(key).get()["Body"].read()
    decoded = read.decode("utf-8")
    assert decoded == _S3_TEXT_BODY


async def test_upload_large_multipart_file(temp_s3_bucket) -> None:
    bytes_to_write = 20 * MiB
    block = b"A" * (4 * MiB)
    key = f"testkey/{uuid4().hex}"
    start = time()
    async with s3_writer_binary(temp_s3_bucket, key) as writer:
        bytes_written = 0
        while bytes_written < bytes_to_write:
            bytes_written += await writer.write(block)
        await writer.flush()
        etag = writer.s3_obj.e_tag
        assert etag.endswith('-3"')
    elapsed = time() - start
    print(elapsed)
    assert elapsed < 2

    async with AsyncS3ObjectReader(s3_object(temp_s3_bucket.name, key)) as reader:
        bytes_read = await reader.read()
        assert len(bytes_read) == bytes_to_write



import asyncio
from asyncio import Task
from typing import Any, Final, Iterator, List, Mapping, Optional, Union

from mypy_boto3_s3.service_resource import MultipartUpload, Object

from nhs_aws_helpers.common import MiB, run_in_executor


class AsyncS3ObjectWriter:
    """provides in-memory buffer to write to S3 objects"""

    MIN_PART_SIZE: Final[int] = 5 * MiB
    DEFAULT_BUFFER_SIZE: Final[int] = 8 * MiB

    def __init__(
        self,
        s3_obj: Object,
        encoding: Optional[str] = None,
        buffer_size: int = DEFAULT_BUFFER_SIZE,
        name: str = "",
        create_object_args: Optional[Mapping[str, Any]] = None,
        max_multipart_concurrency: int = 12,
    ) -> None:
        buffer_size = max(buffer_size, self.MIN_PART_SIZE)
        self._s3_obj = s3_obj
        self._closed = False
        self._encoding = encoding
        self._buffer = b""
        self._bytes_written, self._position = (0, 0)
        self._upload_tasks: List[Task] = []
        self._upload_errors: List[Exception] = []
        self._buffer_size = buffer_size
        self._multipart_upload: Optional[MultipartUpload] = None
        self._name = name
        self._create_object_args = dict(kv for kv in create_object_args.items()) if create_object_args else {}
        self._max_multipart_concurrency = max_multipart_concurrency

    @property
    def s3_obj(self) -> Object:
        return self._s3_obj

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        failed = exc_type is not None
        await self._close(failed=failed)
        if failed:
            return False

    @property
    def mode(self) -> str:
        return "w" if self._encoding else "wb"

    @property
    def name(self) -> str:
        return self._name

    @property
    def closed(self):
        return self._closed

    def fileno(self):
        raise NotImplementedError

    async def flush(self):
        await self._maybe_write()

    @staticmethod
    def seekable() -> bool:
        return False

    @staticmethod
    def readable() -> bool:
        return False

    @staticmethod
    def isatty() -> bool:
        return False

    def tell(self):
        return self._position

    def truncate(self, size=None):
        raise NotImplementedError

    @staticmethod
    def writable() -> bool:
        return True

    def read(self, num_bytes: int = ...) -> Union[str, bytes]:
        raise NotImplementedError

    def readline(self, limit: int = ...) -> Union[str, bytes]:
        raise NotImplementedError

    def readlines(self, hint: int = ...) -> List[Union[str, bytes]]:
        raise NotImplementedError

    def seek(self, offset: int, whence: int = ...) -> int:
        raise NotImplementedError

    def __next__(self) -> Union[str, bytes]:
        raise NotImplementedError

    def __iter__(self) -> Iterator[Union[str, bytes]]:
        raise NotImplementedError

    async def writelines(self, lines: List[Union[str, bytes]]) -> None:
        for arg in lines:
            await self.write(arg)

    async def write(self, data: Union[bytes, str]):
        """Write the given bytes (binary string) to the S3 file.

        There's buffering happening under the covers, so this may not actually
        do any HTTP transfer right away."""

        data = self._maybe_encode(data)
        self._buffer = b"".join([self._buffer, data])  # type: ignore[list-item]

        self._bytes_written += len(data)
        self._position += len(data)

        while self.tell() >= self._buffer_size:
            await self._start_next_part_upload()

        return len(data)

    def _raise_if_errored(self):
        if len(self._upload_errors) < 1:
            return

        raise RuntimeError from self._upload_errors[0]

    async def _start_next_part_upload(self):
        self._raise_if_errored()

        if self._multipart_upload is None:
            args = self._create_object_args or {}
            self._multipart_upload = await run_in_executor(self._s3_obj.initiate_multipart_upload, **args)

        part_data = self._buffer[: self._buffer_size]
        part_num = len(self._upload_tasks) + 1

        self._upload_tasks.append(asyncio.create_task(run_in_executor(self._upload_part, part_num, part_data)))

        self._buffer = self._buffer[self._buffer_size :]
        self._position = len(self._buffer)

    def _upload_part(self, part_num: int, part_data: bytes):
        try:
            assert self._multipart_upload is not None
            self._multipart_upload.Part(part_num).upload(Body=part_data)  # type: ignore[arg-type]
            self._bytes_written += len(part_data)
        except Exception as err:
            self._upload_errors.append(err)

    def _maybe_encode(self, data: Union[str, bytes]) -> bytes:
        if isinstance(data, (bytes, bytearray)):
            return data
        if self._encoding is None:
            return data.encode("utf-8")
        return data.encode(self._encoding)

    async def _maybe_write(self):
        if not self._multipart_upload:
            if not self.tell():
                return

            args = self._create_object_args or {}
            args["Body"] = self._buffer

            await run_in_executor(self._s3_obj.put, **args)
            self._buffer = b""
            self._position = 0
            return

        if self.tell():
            await self._start_next_part_upload()

        if self._upload_tasks:
            await asyncio.gather(*(task for task in self._upload_tasks if not task.done()))

        self._thread_pool = None
        multipart_upload = self._multipart_upload
        self._multipart_upload = None
        if self._upload_errors or self._bytes_written < 1:
            await run_in_executor(multipart_upload.abort)
            self._raise_if_errored()
            return

        result = await run_in_executor(
            multipart_upload.complete,
            MultipartUpload={
                "Parts": [{"ETag": part.e_tag, "PartNumber": part.part_number} for part in multipart_upload.parts.all()]
            },
        )
        assert result

    async def abort(self):
        if self._multipart_upload:
            await run_in_executor(self._multipart_upload.abort)
            for task in self._upload_tasks:
                if task.done():
                    continue
                task.cancel()
        self._closed = True

    async def close(self):
        await self._close()

    async def _close(self, failed: bool = False):
        if self._closed:
            return

        if failed:
            await self.abort()
            return

        await self._maybe_write()

        self._closed = True
